from typing import Any, Optional


class _ErrorMessages(object):
    not_none = '"{0}" should not be None.'
    not_none_or_empty = '"{0}" should not be None or empty.'


class HttpError(Exception):
    """Error from REST API."""

    def __init__(self, message, status_code):
        super(HttpError, self).__init__(message)
        self.status_code = status_code

    def __new__(cls, *args, **kwargs):
        if kwargs.get("status_code") == 401:
            cls = UnauthorizedError
        return Exception.__new__(cls, *args, **kwargs)


class UnauthorizedError(HttpError):
    """Unauthorized error from REST API."""

    def __init__(self, message, status_code):
        message = "Unauthorized, please check your authorization token ({})".format(message)
        super(UnauthorizedError, self).__init__(message, status_code)


def _not_none(param_name: str, param: Optional[Any]):
    if param is None:
        raise TypeError(_ErrorMessages.not_none.format(param_name))


def _not_none_or_empty(param_name: str, param: Optional[str]):
    if param is None:
        raise TypeError(_ErrorMessages.not_none_or_empty.format(param_name))


class ValidationError(Exception):

    def __init__(self, message):
        super(ValidationError, self).__init__(message)



from enum import Enum
from typing import List, Optional, Dict

from pydantic import BaseModel

SYSTEM_RECORD_ID = "system_record_id"
EVAL_SET_INDEX = "eval_set_index"
COUNTRY = "country_iso_code"
RENAMED_INDEX = "index_col"
DEFAULT_INDEX = "index"


class FileColumnMeaningType(Enum):
    MSISDN = "MSISDN"
    EMAIL = "EMAIL"
    IP_ADDRESS = "IP_ADDRESS"
    HEM = "HEM"
    DATE = "DATE"
    DATETIME = "DATETIME"
    SCORE = "SCORE"
    TARGET = "TARGET"
    FEATURE = "FEATURE"
    CUSTOM_KEY = "CUSTOM_KEY"
    COUNTRY = "COUNTRY"
    POSTAL_CODE = "POSTAL_CODE"
    SYSTEM_RECORD_ID = "SYSTEM_RECORD_ID"
    EVAL_SET_INDEX = "EVAL_SET_INDEX"


class SearchKey(Enum):
    EMAIL = FileColumnMeaningType.EMAIL
    HEM = FileColumnMeaningType.HEM
    IP = FileColumnMeaningType.IP_ADDRESS
    PHONE = FileColumnMeaningType.MSISDN
    DATE = FileColumnMeaningType.DATE
    DATETIME = FileColumnMeaningType.DATETIME
    CUSTOM_KEY = FileColumnMeaningType.CUSTOM_KEY
    COUNTRY = FileColumnMeaningType.COUNTRY
    POSTAL_CODE = FileColumnMeaningType.POSTAL_CODE

    @staticmethod
    def personal_keys() -> List["SearchKey"]:
        return [SearchKey.EMAIL, SearchKey.HEM, SearchKey.IP, SearchKey.PHONE]


class DataType(Enum):
    INT = "INT"
    DECIMAL = "DECIMAL"
    DATE_TIME = "DATE_TIME"
    STRING = "STRING"
    BOOLEAN = "BOOLEAN"


class ModelTaskType(Enum):
    BINARY = "BINARY"
    MULTICLASS = "MULTICLASS"
    REGRESSION = "REGRESSION"
    TIMESERIES = "TIMESERIES"


class ModelLabelType(Enum):
    GINI = "gini"
    AUC = "auc"
    RMSE = "rmse"
    ACCURACY = "accuracy"


class RegressionTask(BaseModel):
    mse: float
    rmse: float
    msle: float
    rmsle: float


class BinaryTask(BaseModel):
    auc: float
    gini: float


class MulticlassTask(BaseModel):
    accuracy: float


class FileMetricsInterval(BaseModel):
    date_cut: float
    count: float
    valid_count: float
    avg_target: Optional[float]  # not for multiclass
    avg_score_etalon: Optional[float]


class FileMetrics(BaseModel):
    # etalon metadata
    task_type: Optional[ModelTaskType]
    label: Optional[ModelLabelType]
    count: Optional[int]
    valid_count: Optional[int]
    valid_rate: Optional[float]
    avg_target: Optional[float]
    metrics_binary_etalon: Optional[BinaryTask]
    metrics_regression_etalon: Optional[RegressionTask]
    metrics_multiclass_etalon: Optional[MulticlassTask]
    cuts: Optional[List[float]]
    interval: Optional[List[FileMetricsInterval]]


class NumericInterval(BaseModel):
    minValue: int
    maxValue: int


class FileColumnMetadata(BaseModel):
    index: int
    name: str
    dataType: DataType
    meaningType: FileColumnMeaningType
    minMaxValues: Optional[NumericInterval] = None
    originalName: Optional[str]


class FileMetadata(BaseModel):
    name: str
    description: Optional[str]
    columns: List[FileColumnMetadata]
    searchKeys: List[List[str]]
    hierarchicalGroupKeys: Optional[List[str]]
    hierarchicalSubgroupKeys: Optional[List[str]]
    taskType: Optional[ModelTaskType]
    rowsCount: Optional[int]


class FeaturesFilter(BaseModel):
    minImportance: Optional[float]
    maxPSI: Optional[float]
    maxCount: Optional[int]
    selectedFeatures: Optional[List[str]]


class RuntimeParameters(BaseModel):
    properties: Optional[Dict[str, str]]


class SearchCustomization(BaseModel):
    featuresFilter: Optional[FeaturesFilter]
    extractFeatures: Optional[bool]
    accurateModel: Optional[bool]
    importanceThreshold: Optional[float]
    maxFeatures: Optional[int]
    returnScores: Optional[bool]
    runtimeParameters: Optional[RuntimeParameters]

    def __repr__(self):
        return (
            f"Features filter: {self.featuresFilter}, "
            f"extract features: {self.extractFeatures}, "
            f"accurate model: {self.accurateModel}, "
            f"importance threshold: {self.importanceThreshold}, "
            f"max features: {self.maxFeatures}, "
            f"return scores: {self.returnScores}, "
            f"runtimeParameters: {self.runtimeParameters}"
        )


class CVType(Enum):
    k_fold = "k_fold"
    time_series = "time_series"
    blocked_time_series = "blocked_time_series"



import itertools
import os
import uuid
from copy import deepcopy
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
from pandas.api.types import is_string_dtype
from sklearn.base import TransformerMixin
from sklearn.exceptions import NotFittedError
from sklearn.model_selection import BaseCrossValidator

from upgini.dataset import Dataset
from upgini.http import UPGINI_API_KEY, LoggerFactory
from upgini.mdc import MDC
from upgini.metadata import (
    COUNTRY,
    DEFAULT_INDEX,
    EVAL_SET_INDEX,
    RENAMED_INDEX,
    SYSTEM_RECORD_ID,
    CVType,
    FileColumnMeaningType,
    ModelTaskType,
    RuntimeParameters,
    SearchKey,
)
from upgini.metrics import EstimatorWrapper
from upgini.search_task import SearchTask
from upgini.spinner import Spinner
from upgini.utils.format import Format
from upgini.utils.target_utils import define_task
from upgini.version_validator import validate_version


class FeaturesEnricher(TransformerMixin):
    """Retrieve external features via Upgini that are most relevant to predict your target.

    Parameters
    ----------
    search_keys: dict of str->SearchKey or int->SearchKey
        Dictionary with column names or indices mapping to key types.
        Each of this columns will be used as a search key to find features.

    country_code: str, optional (default=None)
        ISO-3166 COUNTRY code of country for all rows in dataset.

    model_task_type: ModelTaskType, optional (default=None)
        If defined, used as type of training model, else autdefined type will be used

    api_key: str, optional (default=None)
        Token to authorize search requests. You can get it on https://profile.upgini.com/.
        If not specified then read the value from the environment variable UPGINI_API_KEY.

    endpoint: str, optional (default=None)
        URL of Upgini API where search requests are submitted.
        If not specified then used the default value.
        Please don't overwrite it if you are unsure.

    search_id: str, optional (default=None)
        Identifier of fitted enricher.
        If not specified transform could be called only after fit or fit_transform call

    runtime_parameters: dict of str->str, optional (default None).
        Not for public use. Ignore it. It's a way to argument requests with extra parameters.
        Used to trigger experimental features at backend. Used by backend team.

    date_format: str, optional (default=None)
        Format for date column with string type. For example: %Y-%m-%d

    cv: CVType, optional (default=None)
        Type of cross validation: CVType.k_fold, CVType.time_series, CVType.blocked_time_series
        Default cross validation is K-Fold for regressions and Stratified-K-Fold for classifications
    """

    TARGET_NAME = "target"
    RANDOM_STATE = 42

    def __init__(
        self,
        search_keys: Dict[str, SearchKey],
        country_code: Optional[str] = None,
        model_task_type: Optional[ModelTaskType] = None,
        api_key: Optional[str] = None,
        endpoint: Optional[str] = None,
        search_id: Optional[str] = None,
        runtime_parameters: Optional[RuntimeParameters] = None,
        date_format: Optional[str] = None,
        random_state: int = 42,
        cv: Optional[CVType] = None,
    ):
        self.logger = LoggerFactory().get_logger(endpoint, api_key)
        validate_version(self.logger)
        self.search_keys = search_keys
        self.country_code = country_code
        self.__validate_search_keys(search_keys, api_key, search_id)
        self.model_task_type = model_task_type
        self.endpoint = endpoint
        self.api_key = api_key
        self._search_task: Optional[SearchTask] = None
        if search_id:
            search_task = SearchTask(
                search_id,
                endpoint=self.endpoint,
                api_key=self.api_key,
            )
            print("Checking existing search")

            trace_id = str(uuid.uuid4())
            with MDC(trace_id=trace_id):
                try:
                    self._search_task = search_task.poll_result(trace_id, quiet=True)
                    file_metadata = self._search_task.get_file_metadata(trace_id)
                    x_columns = [c.originalName or c.name for c in file_metadata.columns]
                    self.__prepare_feature_importances(trace_id, x_columns)
                    # TODO validate search_keys with search_keys from file_metadata
                    print("Search found. Now you can use transform")
                    self.logger.info(f"FeaturesEnricher successfully initialized with searchTaskId: {search_id}")
                except Exception as e:
                    self.logger.exception("Failed to check existing search")
                    raise e

        self.runtime_parameters = runtime_parameters
        self.date_format = date_format
        self.random_state = random_state
        self.cv = cv
        if cv is not None:
            if self.runtime_parameters is None:
                self.runtime_parameters = RuntimeParameters()
            if self.runtime_parameters.properties is None:
                self.runtime_parameters.properties = {}
            self.runtime_parameters.properties["cv_type"] = cv.name

        self.passed_features: List[str] = []
        self.features_info: pd.DataFrame = pd.DataFrame(columns=["feature_name", "shap_value", "match_percent"])
        self.feature_names_ = []
        self.feature_importances_ = []
        self.enriched_X: Optional[pd.DataFrame] = None
        self.enriched_eval_set: Optional[pd.DataFrame] = None
        self.country_added = False
        self.index_renamed = False

    def fit(
        self,
        X: pd.DataFrame,
        y: Union[pd.Series, np.ndarray, List],
        eval_set: Optional[List[tuple]] = None,
        *,
        calculate_metrics: bool = False,
        estimator: Optional[Any] = None,
        scoring: Union[Callable, str, None] = None,
        importance_threshold: Optional[float] = None,
        max_features: Optional[int] = None,
    ):
        """Fit to data.

        Fits transformer to `X` and `y` with optional parameters `fit_params`.

        Parameters
        ----------
        X : pandas dataframe of shape (n_samples, n_features)
            Input samples.

        y :  array-like of shape (n_samples,) default=None
            Target values.

        eval_set : List[tuple], optional (default=None)
            List of pairs like (X, y) for validation

        keep_input: bool, optional (default=False)
            If True, copy original input columns to the output dataframe.

        importance_threshold: float, optional (default=None)
            Minimum importance shap value for selected features. By default minimum importance is 0.0

        max_features: int, optional (default=None)
            Maximum count of selected most important features. By default it is unlimited

        calculate_metrics : bool (default=False)
            Calculate and show metrics

        estimator : optional (default=None)
            Custom estimator for metrics calculation. It should be sklearn-compatible

        scoring: string, callable or None, optional, default: None
            A string or a scorer callable object / function with signature scorer(estimator, X, y).
            If None, the estimator's score method is used.


        """
        trace_id = str(uuid.uuid4())
        with MDC(trace_id=trace_id):
            self.logger.info(f"Start fit. X shape: {X.shape}. y shape: {len(y)}")
            if eval_set:
                self.logger.info(
                    [
                        f"Eval {i} X shape: {eval_X.shape}, y shape: {len(eval_y)}"
                        for i, (eval_X, eval_y) in enumerate(eval_set)
                    ]
                )

            try:
                self.__inner_fit(
                    trace_id,
                    X,
                    y,
                    eval_set,
                    calculate_metrics=calculate_metrics,
                    estimator=estimator,
                    scoring=scoring,
                    importance_threshold=importance_threshold,
                    max_features=max_features,
                )
                self.logger.info("Fit finished successfully")
            except Exception as e:
                self.logger.exception("Failed inner fit")
                raise e
            finally:
                if self.country_added and COUNTRY in self.search_keys.keys():
                    del self.search_keys[COUNTRY]
                if self.index_renamed and RENAMED_INDEX in self.search_keys.keys():
                    index_key = self.search_keys[RENAMED_INDEX]
                    self.search_keys[DEFAULT_INDEX] = index_key
                    del self.search_keys[RENAMED_INDEX]

    def fit_transform(
        self,
        X: pd.DataFrame,
        y: Union[pd.Series, np.ndarray, List],
        eval_set: Optional[List[tuple]] = None,
        *,
        keep_input: bool = False,
        importance_threshold: Optional[float] = None,
        max_features: Optional[int] = None,
        calculate_metrics: bool = False,
        scoring: Union[Callable, str, None] = None,
        estimator: Optional[Any] = None,
    ) -> pd.DataFrame:
        """Fit to data, then transform it.

        Fits transformer to `X` and `y` with optional parameters `fit_params`
        and returns a transformed version of `X`.
        If keep_input is True, then all input columns are present in output dataframe.

        Parameters
        ----------
        X : pandas dataframe of shape (n_samples, n_features)
            Input samples.

        y :  array-like of shape (n_samples,) default=None
            Target values.

        eval_set : List[tuple], optional (default=None)
            List of pairs like (X, y) for validation

        keep_input: bool, optional (default=False)
            If True, copy original input columns to the output dataframe.

        importance_threshold: float, optional (default=None)
            Minimum importance shap value for selected features. By default minimum importance is 0.0

        max_features: int, optional (default=None)
            Maximum count of selected most important features. By default it is unlimited

        calculate_metrics : bool (default=False)
            Calculate and show metrics

        scoring: string, callable or None, optional, default: None
            A string or a scorer callable object / function with signature scorer(estimator, X, y).
            If None, the estimator's score method is used.

        estimator : optional (default=None)
            Custom estimator for metrics calculation. It should be sklearn-compatible

        Returns
        -------
        X_new : pandas dataframe of shape (n_samples, n_features_new)
            Transformed dataframe, enriched with important features.
        """

        trace_id = str(uuid.uuid4())
        with MDC(trace_id=trace_id):
            self.logger.info(f"Start fit_transform. X shape: {X.shape}. y shape: {len(y)}")
            if eval_set:
                self.logger.info(
                    [
                        f"Eval {i} X shape: {eval_X.shape}, y shape: {len(eval_y)}"
                        for i, (eval_X, eval_y) in enumerate(eval_set)
                    ]
                )
            try:
                self.__inner_fit(
                    trace_id,
                    X,
                    y,
                    eval_set,
                    calculate_metrics=calculate_metrics,
                    scoring=scoring,
                    estimator=estimator,
                    importance_threshold=importance_threshold,
                    max_features=max_features,
                )
                self.logger.info("Fit_transform finished successfully")
            except Exception as e:
                self.logger.exception("Failed in inner_fit")
                raise e
            finally:
                if self.country_added and COUNTRY in self.search_keys.keys():
                    del self.search_keys[COUNTRY]
                if self.index_renamed and RENAMED_INDEX in self.search_keys.keys():
                    index_key = self.search_keys[RENAMED_INDEX]
                    self.search_keys[DEFAULT_INDEX] = index_key
                    del self.search_keys[RENAMED_INDEX]

            return self.transform(
                X, keep_input=keep_input, importance_threshold=importance_threshold, max_features=max_features
            )

    def transform(
        self,
        X: pd.DataFrame,
        *,
        keep_input: bool = False,
        importance_threshold: Optional[float] = None,
        max_features: Optional[int] = None,
    ) -> pd.DataFrame:
        """Transform `X`.

        Returns a transformed version of `X`.
        If keep_input is True, then all input columns are present in output dataframe.

        Parameters
        ----------
        X : pandas dataframe of shape (n_samples, n_features)
            Input samples.

        keep_input: bool, optional (default=False)
            If True, copy original input columns to the output dataframe.

        importance_threshold: float, optional (default=None)
            Minimum importance shap value for selected features. By default minimum importance is 0.0

        max_features: int, optional (default=None)
            Maximum count of selected most important features. By default it is unlimited

        Returns
        -------
        X_new : pandas dataframe of shape (n_samples, n_features_new)
            Transformed dataframe, enriched with important features.
        """

        trace_id = str(uuid.uuid4())
        with MDC(trace_id=trace_id):
            self.logger.info(f"Start transform. X shape: {X.shape}")
            try:
                result = self.__inner_transform(
                    trace_id, X, importance_threshold=importance_threshold, max_features=max_features
                )
                self.logger.info("Transform finished successfully")
            except Exception as e:
                self.logger.exception("Failed to inner transform")
                raise e
            finally:
                if self.country_added and COUNTRY in self.search_keys.keys():
                    del self.search_keys[COUNTRY]
                if self.index_renamed and RENAMED_INDEX in self.search_keys.keys():
                    index_key = self.search_keys[RENAMED_INDEX]
                    self.search_keys["index"] = index_key
                    del self.search_keys[RENAMED_INDEX]

            if keep_input:
                return result
            else:
                return result.drop(columns=[c for c in X.columns if c in result.columns])

    def calculate_metrics(
        self,
        X: pd.DataFrame,
        y: Union[pd.Series, np.ndarray, list],
        eval_set: Optional[List[Tuple[pd.DataFrame, Any]]] = None,
        scoring: Union[Callable, str, None] = None,
        cv: Optional[BaseCrossValidator] = None,
        estimator=None,
        importance_threshold: Optional[float] = None,
        max_features: Optional[int] = None,
        trace_id: Optional[str] = None,
    ) -> Optional[pd.DataFrame]:
        """Calculate metrics

        X : pandas dataframe of shape (n_samples, n_features)
            Input samples same as on the fit.

        y :  array-like of shape (n_samples,) default=None
            Target values same as on the fit.

        eval_set : List[tuple], optional (default=None)
            List of pairs like (X, y) for validation same as on the fit

        scoring: string, callable or None, optional, default: None
            A string or a scorer callable object / function with signature scorer(estimator, X, y).
            If None, the estimator's score method is used.

        cv : BaseCrossValidator, optional (default=None)
            Custom cross validator for metric calculation on train segment

        estimator : sklearn-compatible estimator, optional (default=None)
            Custom estimator for metrics calculation. If not passed then CatBoost will be used

        importance_threshold: float, optional (default=None)
            Minimum importance shap value for selected features. By default minimum importance is 0.0

        max_features: int, optional (default=None)
            Maximum count of selected most important features. By default it is unlimited
        """

        trace_id = trace_id or str(uuid.uuid4())
        with MDC(trace_id=trace_id):
            try:
                if self._search_task is None or self._search_task.initial_max_hit_rate() is None:
                    raise Exception("Fit wasn't completed successfully")
                if self.enriched_X is None:
                    raise Exception("Metrics calculation unavailable after restart. Please run fit again")

                if not isinstance(X, pd.DataFrame):
                    raise Exception(f"Unsupported X type: {type(y)}. Use pandas DataFrame")

                if isinstance(y, np.ndarray) or isinstance(y, list):
                    y = pd.Series(y, name="target")
                elif not isinstance(y, pd.Series):
                    raise Exception(f"Unsupported y type: {type(y)}. Use pandas Series or numpy array or list")

                self.logger.info(f"Calculate metrics with X:\n{X.head(1)}\n and y:\n{y.head(1)}")

                filtered_columns = self.__filtered_columns(
                    X.columns.to_list(), importance_threshold, max_features, only_features=True
                )

                fitting_X = X.drop(columns=[col for col in self.search_keys.keys() if col in X.columns])
                fitting_enriched_X = self.enriched_X[filtered_columns]

                if fitting_X.shape[1] == 0 and fitting_enriched_X.shape[1] == 0:
                    print("WARN: There are no features to calculate metrics")
                    self.logger.warning("No client or relevant ADS features found to calculate metrics")
                    return None

                model_task_type = self.model_task_type or define_task(pd.Series(y), self.logger, silent=True)
                _cv = cv or self.cv

                self.logger.info("Start calculating metrics")
                print("Start calculating metrics")

                with Spinner():
                    # 1 If client features are presented - fit and predict with KFold CatBoost model
                    # on etalon features and calculate baseline metric
                    etalon_metric = None
                    if fitting_X.shape[1] > 0:
                        etalon_metric = EstimatorWrapper.create(
                            estimator, self.logger, model_task_type, _cv, scoring
                        ).cross_val_predict(fitting_X, y)

                    # 2 Fit and predict with KFold Catboost model on enriched tds
                    # and calculate final metric (and uplift)
                    if set(fitting_X.columns) != set(fitting_enriched_X.columns):
                        wrapper = EstimatorWrapper.create(estimator, self.logger, model_task_type, _cv, scoring)
                        enriched_metric = wrapper.cross_val_predict(fitting_enriched_X, y)
                        metric = wrapper.metric_name
                        uplift = None
                        if etalon_metric is not None:
                            uplift = (enriched_metric - etalon_metric) * wrapper.multiplier
                    else:
                        enriched_metric = etalon_metric
                        metric = EstimatorWrapper.create(
                            estimator, self.logger, model_task_type, _cv, scoring
                        ).metric_name
                        uplift = 0.0

                    metrics = [
                        {
                            "segment": "train",
                            "match_rate": self._search_task.initial_max_hit_rate()["value"],  # type: ignore
                            f"baseline {metric}": etalon_metric,
                            f"enriched {metric}": enriched_metric,
                            "uplift": uplift,
                        }
                    ]

                    # 3 If eval_set is presented - fit final model on train enriched data and score each
                    # validation dataset and calculate final metric (and uplift)
                    max_initial_eval_set_metrics = self._search_task.get_max_initial_eval_set_metrics()
                    if eval_set is not None and self.enriched_eval_set is not None:
                        # Fit models
                        etalon_model = None
                        if fitting_X.shape[1] > 0:
                            etalon_model = EstimatorWrapper.create(
                                deepcopy(estimator), self.logger, model_task_type, _cv, scoring
                            )
                            etalon_model.fit(fitting_X, y)

                        if set(fitting_X.columns) != set(fitting_enriched_X.columns):
                            enriched_model = EstimatorWrapper.create(
                                deepcopy(estimator), self.logger, model_task_type, _cv, scoring
                            )
                            enriched_model.fit(fitting_enriched_X, y)
                        elif etalon_model is None:
                            self.logger.error("No client or ADS features, but first validation didn't work")
                            print("There are no features to calculate metrics")
                            return None
                        else:
                            enriched_model = etalon_model

                        for idx, eval_pair in enumerate(eval_set):
                            eval_hit_rate = (
                                max_initial_eval_set_metrics[idx]["hit_rate"] * 100.0
                                if max_initial_eval_set_metrics
                                else None
                            )
                            eval_X = eval_pair[0]
                            eval_X = eval_X.drop(
                                columns=[col for col in self.search_keys.keys() if col in eval_X.columns]
                            )
                            enriched_eval_X = self.enriched_eval_set[self.enriched_eval_set[EVAL_SET_INDEX] == idx + 1]
                            enriched_eval_X = enriched_eval_X[filtered_columns]
                            eval_y = eval_pair[1]

                            etalon_eval_metric = None
                            if etalon_model is not None:
                                etalon_eval_metric = etalon_model.calculate_metric(eval_X, eval_y)

                            enriched_eval_metric = enriched_model.calculate_metric(enriched_eval_X, eval_y)

                            eval_uplift = None
                            if etalon_eval_metric is not None:
                                eval_uplift = (enriched_eval_metric - etalon_eval_metric) * enriched_model.multiplier

                            metrics.append(
                                {
                                    "segment": f"eval {idx + 1}",
                                    "match_rate": eval_hit_rate,
                                    f"baseline {metric}": etalon_eval_metric,
                                    f"enriched {metric}": enriched_eval_metric,
                                    "uplift": eval_uplift,
                                }
                            )
                    self.logger.info("Calculate metrics finished successfully")
                    return pd.DataFrame(metrics).set_index("segment").rename_axis("")
            except Exception as e:
                self.logger.exception("Failed to calculate metrics")
                raise e

    def get_search_id(self) -> Optional[str]:
        """Returns search_id of fitted enricher. It's present only after fit completed"""
        return self._search_task.search_task_id if self._search_task else None

    def get_features_info(self) -> pd.DataFrame:
        """Returns pandas dataframe with importances for each feature"""
        if self._search_task is None or self._search_task.summary is None:
            msg = "Run fit or pass search_id before get features info."
            self.logger.warning(msg)
            raise NotFittedError(msg)

        return self.features_info

    def __inner_transform(
        self,
        trace_id,
        X: pd.DataFrame,
        *,
        importance_threshold: Optional[float],
        max_features: Optional[int],
        silent_mode: bool = False,
    ) -> pd.DataFrame:
        with MDC(trace_id=trace_id):
            if self._search_task is None:
                msg = "`fit` or `fit_transform` should be called before `transform`."
                self.logger.error(msg)
                raise NotFittedError(msg)
            if not isinstance(X, pd.DataFrame):
                msg = f"Only pandas.DataFrame supported for X, but {type(X)} was passed."
                self.logger.error(msg)
                raise TypeError(msg)

            self.__prepare_search_keys(X)

            df = X.copy()

            self.logger.info(f"First dataset row:\n{df.head(1)}")

            df = self.__handle_index_search_keys(df)

            self.__check_string_dates(X)
            df = self.__add_country_code(df)

            meaning_types = {col: key.value for col, key in self.search_keys.items()}
            search_keys = self.__using_search_keys()
            feature_columns = [column for column in df.columns if column not in self.search_keys.keys()]

            df[SYSTEM_RECORD_ID] = [hash(tuple(row)) for row in df[search_keys.keys()].values]  # type: ignore
            meaning_types[SYSTEM_RECORD_ID] = FileColumnMeaningType.SYSTEM_RECORD_ID

            combined_search_keys = []
            for L in range(1, len(search_keys.keys()) + 1):
                for subset in itertools.combinations(search_keys.keys(), L):
                    combined_search_keys.append(subset)

            # Don't pass features in backend on transform
            if len(feature_columns) > 0:
                df_without_features = df.drop(columns=feature_columns)
            else:
                df_without_features = df

            dataset = Dataset(
                "sample_" + str(uuid.uuid4()),
                df=df_without_features,  # type: ignore
                endpoint=self.endpoint,  # type: ignore
                api_key=self.api_key,  # type: ignore
                date_format=self.date_format,  # type: ignore
            )
            dataset.meaning_types = meaning_types
            dataset.search_keys = combined_search_keys
            validation_task = self._search_task.validation(
                trace_id,
                dataset,
                extract_features=True,
                runtime_parameters=self.runtime_parameters,
                silent_mode=silent_mode,
            )

            if not silent_mode:
                print("Executing transform step")
                with Spinner():
                    result, _ = self.__enrich(df, validation_task.get_all_validation_raw_features(trace_id), X.index)
            else:
                result, _ = self.__enrich(df, validation_task.get_all_validation_raw_features(trace_id), X.index)

            filtered_columns = self.__filtered_columns(X.columns.to_list(), importance_threshold, max_features)

            return result[filtered_columns]

    def __validate_search_keys(
        self, search_keys: Dict[str, SearchKey], api_key: Optional[str], search_id: Optional[str]
    ):
        if len(search_keys) == 0:
            if search_id:
                self.logger.error(f"search_id {search_id} provided without search_keys")
                raise ValueError("To transform with search_id please set search_keys to the value used for fitting.")
            else:
                self.logger.error("search_keys not provided")
                raise ValueError("Key columns should be marked up by search_keys.")

        key_types = search_keys.values()

        if SearchKey.DATE in key_types and SearchKey.DATETIME in key_types:
            msg = "Date and datetime search keys are presented simultaniously. Select only one of them"
            self.logger.error(msg)
            raise Exception(msg)

        if SearchKey.EMAIL in key_types and SearchKey.HEM in key_types:
            msg = "Email and HEM search keys are presented simultaniously. Select only one of them"
            self.logger.error(msg)
            raise Exception(msg)

        if SearchKey.POSTAL_CODE in key_types and SearchKey.COUNTRY not in key_types and self.country_code is None:
            msg = "COUNTRY search key should be provided if POSTAL_CODE is presented"
            self.logger.error(msg)
            raise Exception(msg)

        for key_type in SearchKey.__members__.values():
            if key_type != SearchKey.CUSTOM_KEY and len(list(filter(lambda x: x == key_type, key_types))) > 1:
                msg = f"Search key {key_type} presented multiple times"
                self.logger.error(msg)
                raise Exception(msg)

        api_key = api_key or os.environ.get(UPGINI_API_KEY)
        is_registered = api_key is not None and api_key != ""
        non_personal_keys = set(SearchKey.__members__.values()) - set(SearchKey.personal_keys())
        if not is_registered and len(set(key_types).intersection(non_personal_keys)) == 0:
            msg = (
                "Only person-level search keys provided."
                "To run without registration use DATE, COUNTRY and POSTAL_CODE keys"
            )
            self.logger.error(msg + f" Provided search keys: {key_types}")
            raise Exception(msg)

    def __inner_fit(
        self,
        trace_id: str,
        X: pd.DataFrame,
        y: Union[pd.Series, np.ndarray, list, None],
        eval_set: Optional[List[tuple]],
        *,
        calculate_metrics: bool,
        scoring: Union[Callable, str, None],
        estimator: Optional[Any],
        importance_threshold: Optional[float],
        max_features: Optional[int],
    ) -> pd.DataFrame:
        self.enriched_X = None
        if not isinstance(X, pd.DataFrame):
            raise TypeError(f"Only pandas.DataFrame supported for X, but {type(X)} was passed.")
        if not isinstance(y, pd.Series) and not isinstance(y, np.ndarray) and not isinstance(y, list):
            raise TypeError(f"Only pandas.Series or numpy.ndarray or list supported for y, but {type(y)} was passed.")

        if isinstance(y, pd.Series):
            y_array = y.values
        else:
            y_array = y

        if X.shape[0] != len(y_array):
            raise ValueError("X and y should be the same size")

        self.__prepare_search_keys(X)

        df: pd.DataFrame = X.copy()  # type: ignore
        df[self.TARGET_NAME] = y_array

        self.logger.info(f"First dataset row:\n{df.head(1)}")

        df = self.__handle_index_search_keys(df)

        self.__check_string_dates(df)

        model_task_type = self.model_task_type or define_task(df[self.TARGET_NAME], self.logger)

        if eval_set is not None and len(eval_set) > 0:
            df[EVAL_SET_INDEX] = 0
            for idx, eval_pair in enumerate(eval_set):
                if len(eval_pair) != 2:
                    raise TypeError(
                        f"Invalid size of eval_set pair: {len(eval_pair)}. "
                        "It should contain tuples of 2 elements: X and y."
                    )
                eval_X = eval_pair[0]
                eval_y = eval_pair[1]
                if not isinstance(eval_X, pd.DataFrame):
                    raise TypeError(
                        f"Only pandas.DataFrame supported for X in eval_set, but {type(eval_X)} was passed."
                    )
                if eval_X.columns.to_list() != X.columns.to_list():
                    raise Exception("Eval set columns differ to X columns")
                if (
                    not isinstance(eval_y, pd.Series)
                    and not isinstance(eval_y, np.ndarray)
                    and not isinstance(eval_y, list)
                ):
                    raise TypeError(
                        "pandas.Series or numpy.ndarray or list supported for y in eval_set, "
                        f"but {type(eval_y)} was passed."
                    )
                eval_df: pd.DataFrame = eval_X.copy()  # type: ignore
                eval_df[self.TARGET_NAME] = pd.Series(eval_y)
                eval_df[EVAL_SET_INDEX] = idx + 1
                df = pd.concat([df, eval_df], ignore_index=True)

        df = self.__add_country_code(df)

        non_feature_columns = [self.TARGET_NAME, EVAL_SET_INDEX] + list(self.search_keys.keys())
        meaning_types = {
            **{col: key.value for col, key in self.search_keys.items()},
            **{str(c): FileColumnMeaningType.FEATURE for c in df.columns if c not in non_feature_columns},
        }
        meaning_types[self.TARGET_NAME] = FileColumnMeaningType.TARGET
        if eval_set is not None and len(eval_set) > 0:
            meaning_types[EVAL_SET_INDEX] = FileColumnMeaningType.EVAL_SET_INDEX

        search_keys = self.__using_search_keys()

        df = self.__add_fit_system_record_id(df, meaning_types)

        combined_search_keys = []
        for L in range(1, len(search_keys.keys()) + 1):
            for subset in itertools.combinations(search_keys.keys(), L):
                combined_search_keys.append(subset)

        dataset = Dataset(
            "tds_" + str(uuid.uuid4()),
            df=df,  # type: ignore
            model_task_type=model_task_type,  # type: ignore
            endpoint=self.endpoint,  # type: ignore
            api_key=self.api_key,  # type: ignore
            date_format=self.date_format,  # type: ignore
            random_state=self.random_state,  # type: ignore
        )
        dataset.meaning_types = meaning_types
        dataset.search_keys = combined_search_keys

        self.passed_features = [
            column for column, meaning_type in meaning_types.items() if meaning_type == FileColumnMeaningType.FEATURE
        ]

        self._search_task = dataset.search(
            trace_id,
            extract_features=True,
            runtime_parameters=self.runtime_parameters,
        )

        self.__prepare_feature_importances(trace_id, list(X.columns))

        self.__show_selected_features()

        try:
            self.enriched_X, self.enriched_eval_set = self.__enrich(
                df, self._search_task.get_all_initial_raw_features(trace_id), X.index
            )
        except Exception as e:
            self.logger.exception("Failed to download features")
            raise e

        if calculate_metrics:
            self.__show_metrics(X, y, eval_set, scoring, estimator, importance_threshold, max_features, trace_id)

        filtered_columns = self.__filtered_columns(X.columns.to_list(), importance_threshold, max_features)

        return self.enriched_X[filtered_columns]

    def __handle_index_search_keys(self, df: pd.DataFrame) -> pd.DataFrame:
        index_names = df.index.names if df.index.names != [None] else [DEFAULT_INDEX]
        index_search_keys = set(index_names).intersection(self.search_keys.keys())
        if len(index_search_keys) > 0:
            for index_name in index_search_keys:
                if index_name not in df.columns:
                    if df.index.names == [None]:
                        df[index_name] = df.index
                    else:
                        df[index_name] = df.index.get_level_values(index_name)
            df = df.reset_index(drop=True)
            if DEFAULT_INDEX in index_names:
                df = df.rename(columns={DEFAULT_INDEX: RENAMED_INDEX})
                self.search_keys[RENAMED_INDEX] = self.search_keys[DEFAULT_INDEX]
                del self.search_keys[DEFAULT_INDEX]
                self.index_renamed = True
        else:
            df = df.reset_index(drop=True)
        return df

    def __using_search_keys(self) -> Dict[str, SearchKey]:
        return {col: key for col, key in self.search_keys.items() if key != SearchKey.CUSTOM_KEY}

    def __is_date_key_present(self) -> bool:
        return len({SearchKey.DATE, SearchKey.DATETIME}.intersection(self.search_keys.values())) != 0

    def __add_fit_system_record_id(
        self, df: pd.DataFrame, meaning_types: Dict[str, FileColumnMeaningType]
    ) -> pd.DataFrame:
        if (self.cv is None or self.cv == CVType.k_fold) and self.__is_date_key_present():
            date_column = [
                col
                for col, t in meaning_types.items()
                if t in [FileColumnMeaningType.DATE, FileColumnMeaningType.DATETIME]
            ]
            df.sort_values(by=date_column, kind="mergesort")
            pass
        df = df.reset_index(drop=True)
        df = df.reset_index()
        df = df.rename(columns={DEFAULT_INDEX: SYSTEM_RECORD_ID})
        meaning_types[SYSTEM_RECORD_ID] = FileColumnMeaningType.SYSTEM_RECORD_ID
        return df

    def __check_string_dates(self, df: pd.DataFrame):
        for column, search_key in self.search_keys.items():
            if search_key in [SearchKey.DATE, SearchKey.DATETIME] and is_string_dtype(df[column]):
                if self.date_format is None or len(self.date_format) == 0:
                    msg = (
                        f"Date column `{column}` has string type, but constructor argument `date_format` is empty.\n"
                        "Please, convert column to datetime type or pass date format implicitly"
                    )
                    self.logger.error(msg)
                    raise Exception(msg)

    def __add_country_code(self, df: pd.DataFrame) -> pd.DataFrame:
        if self.country_code is not None and SearchKey.COUNTRY not in self.search_keys.values():
            self.logger.info(f"Add COUNTRY column with {self.country_code} value")
            df[COUNTRY] = self.country_code
            self.search_keys[COUNTRY] = SearchKey.COUNTRY
            self.country_added = True
        return df

    def __enrich(
        self,
        df: pd.DataFrame,
        result_features: Optional[pd.DataFrame],
        original_index: pd.Index,
    ) -> Tuple[pd.DataFrame, Optional[pd.DataFrame]]:
        if result_features is None:
            self.logger.error(f"result features not found by search_task_id: {self.get_search_id()}")
            raise RuntimeError("Search engine crashed on this request.")
        result_features = (
            result_features.drop(columns=EVAL_SET_INDEX)
            if EVAL_SET_INDEX in result_features.columns
            else result_features
        )
        df_without_target = df.drop(columns=self.TARGET_NAME) if self.TARGET_NAME in df.columns else df
        result = pd.merge(
            df_without_target,
            result_features,
            left_on=SYSTEM_RECORD_ID,
            right_on=SYSTEM_RECORD_ID,
            how="left",
            # suffixes=("_ads", "")
        )

        if EVAL_SET_INDEX in result.columns:
            result_train = result[result[EVAL_SET_INDEX] == 0]
            result_eval_set = result[result[EVAL_SET_INDEX] != 0]
            result_train = result_train.drop(columns=EVAL_SET_INDEX)
        else:
            result_train = result
            result_eval_set = None

        result_train.index = original_index
        if SYSTEM_RECORD_ID in result.columns:
            result_train = result_train.drop(columns=SYSTEM_RECORD_ID)
            if result_eval_set is not None:
                result_eval_set = result_eval_set.drop(columns=SYSTEM_RECORD_ID)

        return result_train, result_eval_set

    def __prepare_feature_importances(self, trace_id: str, x_columns: List[str]):
        if self._search_task is None:
            raise NotFittedError("`fit` or `fit_transform` should be called before `transform`.")
        importances = self._search_task.initial_features(trace_id)

        def feature_metadata_by_name(name: str):
            for f in importances:
                if f["feature_name"] == name:
                    return f

        self.feature_names_ = []
        self.feature_importances_ = []
        features_info = []

        service_columns = [SYSTEM_RECORD_ID, EVAL_SET_INDEX, self.TARGET_NAME]

        importances.sort(key=lambda m: -m["shap_value"])
        for feature_metadata in importances:
            if feature_metadata["feature_name"] not in x_columns:
                self.feature_names_.append(feature_metadata["feature_name"])
                self.feature_importances_.append(feature_metadata["shap_value"])
            features_info.append(feature_metadata)

        for x_column in x_columns:
            if x_column in (list(self.search_keys.keys()) + service_columns):
                continue
            feature_metadata = feature_metadata_by_name(x_column)
            if feature_metadata is None:
                features_info.append(
                    {
                        "feature_name": x_column,
                        "shap_value": 0.0,
                        "coverage %": None,  # TODO fill from X
                    }
                )

        if len(features_info) > 0:
            self.features_info = pd.DataFrame(features_info)

    def __filtered_importance_names(
        self, importance_threshold: Optional[float], max_features: Optional[int]
    ) -> List[str]:
        if len(self.feature_names_) == 0:
            return []

        filtered_importances = list(zip(self.feature_names_, self.feature_importances_))
        # temporary workaround. generate this column later
        filtered_importances = [
            (name, importance) for name, importance in filtered_importances if name != "email_domain"
        ]
        if importance_threshold is not None:
            filtered_importances = [
                (name, importance) for name, importance in filtered_importances if importance > importance_threshold
            ]
        if max_features is not None:
            filtered_importances = list(filtered_importances)[:max_features]
        if len(filtered_importances) == 0:
            return []
        filtered_importance_names, _ = zip(*filtered_importances)
        return list(filtered_importance_names)

    def __prepare_search_keys(self, x: pd.DataFrame):
        valid_search_keys = {}
        api_key = self.api_key or os.environ.get(UPGINI_API_KEY)
        is_registered = api_key is not None and api_key != ""
        for column_id, meaning_type in self.search_keys.items():
            column_name = None
            if isinstance(column_id, str):
                column_name = column_id
                valid_search_keys[column_name] = meaning_type
            elif isinstance(column_id, int):
                column_name = x.columns[column_id]
                valid_search_keys[column_name] = meaning_type
            else:
                msg = f"Unsupported type of key in search_keys: {type(column_id)}."
                self.logger.error(msg)
                raise ValueError(msg)

            if meaning_type == SearchKey.COUNTRY and self.country_code is not None:
                msg = (
                    "SearchKey.COUNTRY cannot be used together with a iso_code property at the same time. "
                    "Define only one"
                )
                self.logger.error(msg)
                raise ValueError(msg)

            if not is_registered and meaning_type in SearchKey.personal_keys():
                msg = f"Search key {meaning_type} not available without registration. It will be ignored"
                self.logger.warning(msg)
                print("WARNING: " + msg)
                valid_search_keys[column_name] = SearchKey.CUSTOM_KEY

        self.search_keys = valid_search_keys

    def __show_metrics(
        self,
        X: pd.DataFrame,
        y: Union[pd.Series, np.ndarray, list],
        eval_set: Optional[List[Tuple[pd.DataFrame, Any]]],
        scoring: Union[Callable, str, None],
        estimator: Optional[Any],
        importance_threshold: Optional[float],
        max_features: Optional[int],
        trace_id: str,
    ):
        metrics = self.calculate_metrics(
            X,
            y,
            eval_set,
            scoring=scoring,
            estimator=estimator,
            importance_threshold=importance_threshold,
            max_features=max_features,
            trace_id=trace_id,
        )
        if metrics is not None:
            msg = "\nQuality metrics"

            try:
                from IPython.display import display

                print(Format.GREEN + Format.BOLD + msg + Format.END)
                display(metrics)
            except ImportError:
                print(msg)
                print(metrics)

    def __show_selected_features(self):
        search_keys = self.__using_search_keys().keys()
        msg = f"\n{len(self.feature_names_)} relevant feature(s) found with the search keys: {list(search_keys)}"

        try:
            from IPython.display import display

            print(Format.GREEN + Format.BOLD + msg + Format.END)
            display(self.features_info.head(60))
        except ImportError:
            print(msg)
            print(self.features_info.head(60))

    def __is_quality_by_metrics_low(self) -> bool:
        if self._search_task is None:
            return False
        if len(self.passed_features) > 0 and self._search_task.task_type is not None:
            max_uplift = self._search_task.initial_max_uplift()
            if self._search_task.task_type == ModelTaskType.BINARY:
                threshold = 0.002
            elif self._search_task.task_type == ModelTaskType.MULTICLASS:
                threshold = 3.0
            elif self._search_task.task_type == ModelTaskType.REGRESSION:
                threshold = 0.0
            else:
                return False
            if max_uplift is not None and max_uplift["value"] < threshold:
                return True
        elif self._search_task.task_type is not None:
            max_auc = self._search_task.initial_max_auc()
            if self._search_task.task_type == ModelTaskType.BINARY and max_auc is not None:
                if max_auc["value"] < 0.55:
                    return True
        return False

    def __validate_importance_threshold(self, importance_threshold: Optional[float]) -> float:
        try:
            return float(importance_threshold) if importance_threshold is not None else 0.0
        except ValueError:
            self.logger.exception(f"Invalid importance_threshold provided: {importance_threshold}")
            raise ValueError("importance_threshold should be float")

    def __validate_max_features(self, max_features: Optional[int]) -> int:
        try:
            return int(max_features) if max_features is not None else 400
        except ValueError:
            self.logger.exception(f"Invalid max_features provided: {max_features}")
            raise ValueError("max_features should be int")

    def __filtered_columns(
        self,
        x_columns: List[str],
        importance_threshold: Optional[float],
        max_features: Optional[int],
        only_features: bool = False,
    ) -> List[str]:
        importance_threshold = self.__validate_importance_threshold(importance_threshold)
        max_features = self.__validate_max_features(max_features)

        exclude_columns = list(self.search_keys.keys()) if only_features else []

        return [col for col in x_columns if col not in exclude_columns] + self.__filtered_importance_names(
            importance_threshold, max_features
        )

    def __check_quality(self, no_data_found: bool):
        if no_data_found or self.__is_quality_by_metrics_low():
            try:
                from IPython.display import HTML, display  # type: ignore

                display(
                    HTML(
                        "<h9>Oops, looks like we're not able to find data which gives a strong uplift for your ML "
                        "algorithm.<br>If you have ANY data which you might consider as royalty and "
                        "license-free and potentially valuable for supervised ML applications,<br> we shall be "
                        "happy to give you free individual access in exchange for sharing this data with "
                        "community.<br>Just upload your data sample right from Jupyter. We will check your data "
                        "sharing proposal and get back to you ASAP."
                    )
                )
                display(
                    HTML(
                        "<a href='https://github.com/upgini/upgini/blob/main/README.md"
                        "#share-license-free-data-with-community' "
                        "target='_blank'>How to upload your data sample from Jupyter</a>"
                    )
                )
            except ImportError:
                print("Oops, looks like we're not able to find data which gives a strong uplift for your ML algorithm.")
                print(
                    "If you have ANY data which you might consider as royalty and license-free and potentially "
                    "valuable for supervised ML applications,"
                )
                print(
                    "we shall be happy to give you free individual access in exchange for sharing this data with "
                    "community."
                )
                print(
                    "Just upload your data sample right from Jupyter. We will check your data sharing proposal and "
                    "get back to you ASAP."
                )
                print("https://github.com/upgini/upgini/blob/main/README.md#share-license-free-data-with-community")



import pandas as pd
from pandas.api.types import is_numeric_dtype
import numpy as np
import logging

from upgini.metadata import ModelTaskType


def define_task(y: pd.Series, logger: logging.Logger, silent: bool = False) -> ModelTaskType:
    target = y.dropna()
    if is_numeric_dtype(target):
        target = target.loc[np.isfinite(target)]  # type: ignore
    else:
        target = target.loc[target != ""]
    if len(target) == 0:
        raise ValueError("Target is empty in all rows")
    target_items = target.nunique()
    target_ratio = target_items / len(target)
    if (target_items > 50 or (target_items > 2 and target_ratio > 0.2)) and is_numeric_dtype(target):
        task = ModelTaskType.REGRESSION
    elif target_items <= 2:
        if is_numeric_dtype(target):
            task = ModelTaskType.BINARY
        else:
            raise ValueError("Binary target should be numerical")
    else:
        task = ModelTaskType.MULTICLASS
    logger.info(f"Detected task type: {task}")
    if not silent:
        print(f"Detected task type: {task}")
    return task



from typing import Optional

import pandas as pd
from pandas.api.types import is_float_dtype, is_int64_dtype, is_string_dtype


def phone_to_int(df: pd.DataFrame, phone_column_name: str) -> pd.DataFrame:
    """
    Convention: phone number is always presented as int number.
    phone_number = Country code + National Destination Code + Subscriber Number.
    Examples:
    41793834315     for Switzerland
    46767040672     for Sweden
    861065529988    for China
    18143008198     for the USA
    Inplace conversion of phone to int.

    Method will remove all non numeric chars from string and convert it to int.
    None will be set for phone numbers that couldn't be converted to int
    """
    if is_string_dtype(df[phone_column_name]):
        df[phone_column_name] = df[phone_column_name].apply(phone_str_to_int_safe)
    elif is_float_dtype(df[phone_column_name]):
        df[phone_column_name] = df[phone_column_name].apply(phone_float_to_int_safe)
    elif is_int64_dtype(df[phone_column_name]):
        df[phone_column_name] = df[phone_column_name].apply(phone_int_to_int_safe)
    else:
        raise Exception(
            f"phone_column_name {phone_column_name} doesn't have supported dtype. Dataset dtypes: {df.dtypes}. "
            f"Contact developer and request to implement conversion of {phone_column_name} to int"
        )
    return df


def phone_float_to_int_safe(value: float) -> Optional[int]:
    try:
        return validate_length(int(value))
    except Exception:
        return None


def phone_int_to_int_safe(value: float) -> Optional[int]:
    try:
        return validate_length(int(value))
    except Exception:
        return None


def phone_str_to_int_safe(value: str) -> Optional[int]:
    try:
        numeric_filter = filter(str.isdigit, value)
        numeric_string = "".join(numeric_filter)
        return validate_length(int(numeric_string))
    except Exception:
        return None


def validate_length(value: int) -> Optional[int]:
    if value < 10000000 or value > 999999999999999:
        return None
    else:
        return value



import csv
import os
import tempfile
import time
from hashlib import sha256
from ipaddress import IPv4Address, ip_address
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
from upgini.sampler.random_under_sampler import RandomUnderSampler
from pandas.api.types import is_bool_dtype as is_bool
from pandas.api.types import is_datetime64_any_dtype as is_datetime
from pandas.api.types import (
    is_float_dtype,
    is_integer_dtype,
    is_numeric_dtype,
    is_string_dtype,
)
from pandas.core.dtypes.common import is_period_dtype

from upgini.errors import ValidationError
from upgini.http import UPGINI_API_KEY, LoggerFactory, get_rest_client
from upgini.metadata import (
    EVAL_SET_INDEX,
    SYSTEM_RECORD_ID,
    DataType,
    FeaturesFilter,
    FileColumnMeaningType,
    FileColumnMetadata,
    FileMetadata,
    FileMetrics,
    ModelTaskType,
    NumericInterval,
    RuntimeParameters,
    SearchCustomization,
)
from upgini.normalizer.phone_normalizer import phone_to_int
from upgini.search_task import SearchTask


class Dataset(pd.DataFrame):
    MIN_ROWS_COUNT = 100
    MAX_ROWS_REGISTERED = 299_999
    MAX_ROWS_UNREGISTERED = 149_999
    FIT_SAMPLE_ROWS = 100_000
    FIT_SAMPLE_THRESHOLD = FIT_SAMPLE_ROWS * 3
    IMBALANCE_THESHOLD = 0.4
    MIN_TARGET_CLASS_COUNT = 100
    MAX_MULTICLASS_CLASS_COUNT = 100
    MIN_SUPPORTED_DATE_TS = 946684800000  # 2000-01-01
    MAX_FEATURES_COUNT = 1100

    _metadata = [
        "dataset_name",
        "description",
        "meaning_types",
        "search_keys",
        "ignore_columns",
        "hierarchical_group_keys",
        "hierarchical_subgroup_keys",
        "date_format",
        "random_state",
        "task_type",
        "initial_data",
        "file_upload_id",
        "etalon_def",
        "endpoint",
        "api_key",
        "columns_renaming",
        "sampled",
    ]

    def __init__(
        self,
        dataset_name: str,
        description: Optional[str] = None,
        df: Optional[pd.DataFrame] = None,
        path: Optional[str] = None,
        meaning_types: Optional[Dict[str, FileColumnMeaningType]] = None,
        search_keys: Optional[List[Tuple[str, ...]]] = None,
        model_task_type: Optional[ModelTaskType] = None,
        date_format: Optional[str] = None,
        random_state: Optional[int] = None,
        endpoint: Optional[str] = None,
        api_key: Optional[str] = None,
        **kwargs,
    ):
        if df is not None:
            data = df.copy()
        elif path is not None:
            if "sep" in kwargs:
                data = pd.read_csv(path, **kwargs)
            else:
                # try different separators: , ; \t ...
                with open(path, mode="r") as csvfile:
                    sep = csv.Sniffer().sniff(csvfile.read(2048)).delimiter
                kwargs["sep"] = sep
                data = pd.read_csv(path, **kwargs)
        else:
            raise ValueError("DataFrame or path to file should be passed")
        if isinstance(data, pd.DataFrame):
            super(Dataset, self).__init__(data)  # type: ignore
        else:
            raise ValueError("Iteration is not supported. Remove `iterator` and `chunksize` arguments and try again")

        self.dataset_name = dataset_name
        self.task_type = model_task_type
        self.description = description
        self.meaning_types = meaning_types
        self.search_keys = search_keys
        self.ignore_columns = []
        self.hierarchical_group_keys = []
        self.hierarchical_subgroup_keys = []
        self.date_format = date_format
        self.initial_data = data.copy()
        self.file_upload_id: Optional[str] = None
        self.etalon_def: Optional[Dict[str, str]] = None
        self.endpoint = endpoint
        self.api_key = api_key
        self.random_state = random_state
        self.columns_renaming: Dict[str, str] = {}
        self.sampled: bool = False
        self.logger = LoggerFactory().get_logger(endpoint, api_key)

    @property
    def meaning_types_checked(self) -> Dict[str, FileColumnMeaningType]:
        if self.meaning_types is None:
            raise ValueError("meaning_types is empty")
        else:
            return self.meaning_types

    @property
    def search_keys_checked(self) -> List[Tuple[str, ...]]:
        if self.search_keys is None:
            raise ValueError("search_keys is empty")
        else:
            return self.search_keys

    @property
    def etalon_def_checked(self) -> Dict[str, str]:
        if self.etalon_def is None:
            self.etalon_def = {
                v.value: k for k, v in self.meaning_types_checked.items() if v != FileColumnMeaningType.FEATURE
            }

        return self.etalon_def

    def __validate_min_rows_count(self):
        if self.shape[0] < self.MIN_ROWS_COUNT:
            raise ValueError(f"X should contain at least {self.MIN_ROWS_COUNT} valid distinct rows")

    def __validate_max_row_count(self):
        api_key = self.api_key or os.environ.get(UPGINI_API_KEY)
        is_registered = api_key is not None and api_key != ""
        if is_registered:
            if len(self) > self.MAX_ROWS_REGISTERED:
                raise ValueError(
                    f"Total X + eval_set rows count limit is {self.MAX_ROWS_REGISTERED}. "
                    "Please sample X and eval_set"
                )
        else:
            if len(self) > self.MAX_ROWS_UNREGISTERED:
                raise ValueError(
                    f"For unregistered users total rows count limit for X + eval_set is {self.MAX_ROWS_UNREGISTERED}. "
                    "Please register to increase the limit"
                )

    def __rename_columns(self):
        # self.logger.info("Replace restricted symbols in column names")
        for column in self.columns:
            if len(column) == 0:
                raise ValueError("Some of column names are empty. Add names and try again, please")
            new_column = str(column).lower()
            if ord(new_column[0]) not in range(ord("a"), ord("z")):
                new_column = "a" + new_column
            for idx, c in enumerate(new_column):
                if ord(c) not in range(ord("a"), ord("z")) and ord(c) not in range(ord("0"), ord("9")):
                    new_column = new_column[:idx] + "_" + new_column[idx + 1 :]
            self.rename(columns={column: new_column}, inplace=True)
            self.meaning_types = {
                (new_column if key == str(column) else key): value for key, value in self.meaning_types_checked.items()
            }
            self.search_keys = [
                tuple(new_column if key == str(column) else key for key in keys) for keys in self.search_keys_checked
            ]
            self.columns_renaming[new_column] = str(column)

    def __validate_too_long_string_values(self):
        """Check that string values less than 400 characters"""
        # self.logger.info("Validate too long string values")
        for col in self.columns:
            if is_string_dtype(self[col]):
                max_length: int = self[col].astype("str").str.len().max()
                if max_length > 400:
                    raise ValueError(
                        f"Columns {col} are too long: {max_length} characters. "
                        "Remove this columns or trim length to 50 characters"
                    )

    def __clean_duplicates(self):
        """Clean DataSet from full duplicates."""
        # self.logger.info("Clean full duplicates")
        nrows = len(self)
        # Remove absolute duplicates (exclude system_record_id)
        unique_columns = self.columns.tolist()
        unique_columns.remove(SYSTEM_RECORD_ID)
        self.logger.info(f"Dataset shape before clean duplicates: {self.shape}")
        self.drop_duplicates(subset=unique_columns, inplace=True)
        self.logger.info(f"Dataset shape after clean duplicates: {self.shape}")
        nrows_after_full_dedup = len(self)
        share_full_dedup = 100 * (1 - nrows_after_full_dedup / nrows)
        if share_full_dedup > 0:
            print(f"{share_full_dedup:.5f}% of the rows are fully duplicated")
        target_column = self.etalon_def_checked.get(FileColumnMeaningType.TARGET.value)
        if target_column is not None:
            unique_columns.remove(target_column)
            # unique_columns.remove(SYSTEM_RECORD_ID)
            self.drop_duplicates(subset=unique_columns, inplace=True)
            nrows_after_tgt_dedup = len(self)
            share_tgt_dedup = 100 * (1 - nrows_after_tgt_dedup / nrows_after_full_dedup)
            if nrows_after_tgt_dedup < nrows_after_full_dedup:
                msg = (
                    f"{share_tgt_dedup:.5f}% of rows in X are duplicates with different y values. "
                    "Please check X dataframe"
                )
                self.logger.error(msg)
                raise ValueError(msg)

    def __convert_bools(self):
        """Convert bool columns True -> 1, False -> 0"""
        # self.logger.info("Converting bool to int")
        for col in self.columns:
            if is_bool(self[col]):
                self[col] = self[col].astype("Int64")

    def __convert_float16(self):
        """Convert float16 to float"""
        # self.logger.info("Converting float16 to float")
        for col in self.columns:
            if is_float_dtype(self[col]):
                self[col] = self[col].astype("float64")

    def __correct_decimal_comma(self):
        """Check DataSet for decimal commas and fix them"""
        # self.logger.info("Correct decimal commas")
        tmp = self.head(10)
        # all columns with sep="," will have dtype == 'object', i.e string
        # sep="." will be casted to numeric automatically
        cls_to_check = [i for i in tmp.columns if is_string_dtype(tmp[i])]
        for col in cls_to_check:
            if tmp[col].astype(str).str.match("^[0-9]+,[0-9]*$").any():
                self[col] = self[col].astype(str).str.replace(",", ".").astype(np.float64)

    def __to_millis(self):
        """Parse date column and transform it to millis"""
        date = self.etalon_def_checked.get(FileColumnMeaningType.DATE.value) or self.etalon_def_checked.get(
            FileColumnMeaningType.DATETIME.value
        )

        def intToOpt(i: int) -> Optional[int]:
            if i == -9223372036855:
                return None
            else:
                return i

        if date is not None and date in self.columns:
            # self.logger.info("Transform date column to millis")
            if is_string_dtype(self[date]):
                self[date] = (
                    pd.to_datetime(self[date], format=self.date_format).dt.floor("D").view(np.int64) // 1_000_000
                )
            elif is_datetime(self[date]):
                self[date] = self[date].dt.floor("D").view(np.int64) // 1_000_000
            elif is_period_dtype(self[date]):
                self[date] = pd.to_datetime(self[date].astype("string")).dt.floor("D").view(np.int64) // 1_000_000
            elif is_numeric_dtype(self[date]):
                msg = f"Unsupported type of date column {date}. Convert to datetime please."
                self.logger.error(msg)
                raise Exception(msg)

            self[date] = self[date].apply(lambda x: intToOpt(x)).astype("Int64")

    @staticmethod
    def __email_to_hem(email: str) -> Optional[str]:
        if email is None or not isinstance(email, str) or email == "":
            return None
        else:
            return sha256(email.lower().encode("utf-8")).hexdigest()

    def __hash_email(self):
        """Add column with HEM if email presented in search keys"""
        email = self.etalon_def_checked.get(FileColumnMeaningType.EMAIL.value)
        if email is not None and email in self.columns:
            # self.logger.info("Hashing email")
            generated_hem_name = "generated_hem"
            self[generated_hem_name] = self[email].apply(self.__email_to_hem)
            self.meaning_types_checked[generated_hem_name] = FileColumnMeaningType.HEM
            self.meaning_types_checked.pop(email)
            self.etalon_def_checked[FileColumnMeaningType.HEM.value] = generated_hem_name
            del self.etalon_def_checked[FileColumnMeaningType.EMAIL.value]
            self.search_keys = [
                tuple(key if key != email else generated_hem_name for key in search_group)
                for search_group in self.search_keys_checked
            ]
            self["email_domain"] = self[email].str.split("@").str[1]
            self.drop(columns=email, inplace=True)

    @staticmethod
    def __ip_to_int(ip: Union[str, int, IPv4Address]) -> Optional[int]:
        try:
            return int(ip_address(ip))
        except Exception:
            return None

    def __convert_ip(self):
        """Convert ip address to int"""
        ip = self.etalon_def_checked.get(FileColumnMeaningType.IP_ADDRESS.value)
        if ip is not None and ip in self.columns:
            # self.logger.info("Convert ip address to int")
            self[ip] = self[ip].apply(self.__ip_to_int).astype("Int64")

    def __normalize_iso_code(self):
        iso_code = self.etalon_def_checked.get(FileColumnMeaningType.COUNTRY.value)
        if iso_code is not None and iso_code in self.columns:
            # self.logger.info("Normalize iso code column")
            self[iso_code] = (
                self[iso_code]
                .astype(str)
                .str.upper()
                .str.replace(r"[^A-Z]", "", regex=True)
                .str.replace("UK", "GB", regex=False)
            )

    def __normalize_postal_code(self):
        postal_code = self.etalon_def_checked.get(FileColumnMeaningType.POSTAL_CODE.value)
        if postal_code is not None and postal_code in self.columns:
            # self.logger.info("Normalize postal code")
            self[postal_code] = (
                self[postal_code]
                .astype(str)
                .str.upper()
                .str.replace(r"[^0-9A-Z]", "", regex=True)
                .str.replace(r"^0+\B", "", regex=True)
            )

    def __remove_old_dates(self):
        date_column = self.etalon_def_checked.get(FileColumnMeaningType.DATE.value) or self.etalon_def_checked.get(
            FileColumnMeaningType.DATETIME.value
        )
        if date_column is not None:
            old_subset = self[self[date_column] < self.MIN_SUPPORTED_DATE_TS]
            if len(old_subset) > 0:
                self.logger.info(f"df before dropping old rows: {self.shape}")
                self.drop(index=old_subset.index, inplace=True)
                self.logger.info(f"df after dropping old rows: {self.shape}")
                msg = "We don't have data before '2000-01-01' and removed all earlier records from the search dataset"
                self.logger.warning(msg)
                print("WARN: ", msg)
                if len(self) == 0:
                    raise Exception("There is empty train dataset after dropping old rows")

    def __drop_ignore_columns(self):
        """Drop ignore columns"""
        columns_to_drop = list(set(self.columns) & set(self.ignore_columns))
        if len(columns_to_drop) > 0:
            # self.logger.info(f"Dropping ignore columns: {self.ignore_columns}")
            self.drop(columns_to_drop, axis=1, inplace=True)

    def __target_value(self) -> pd.Series:
        target_column = self.etalon_def_checked.get(FileColumnMeaningType.TARGET.value, "")
        target: pd.Series = self[target_column]
        # clean target from nulls
        target.dropna(inplace=True)
        if is_numeric_dtype(target):
            target = target.loc[np.isfinite(target)]  # type: ignore
        else:
            target = target.loc[target != ""]

        return target

    def __validate_target(self):
        # self.logger.info("Validating target")
        target_column = self.etalon_def_checked.get(FileColumnMeaningType.TARGET.value, "")
        target = self[target_column]

        if self.task_type == ModelTaskType.BINARY:
            if not is_integer_dtype(target):
                try:
                    self[target_column] = self[target_column].astype("int")
                except ValueError:
                    self.logger.exception("Failed to cast target to integer for binary task type")
                    raise ValidationError(
                        f"Unexpected dtype of target for binary task type: {target.dtype}." " Expected int or bool"
                    )
            target_classes_count = target.nunique()
            if target_classes_count != 2:
                msg = f"Binary task type should contain only 2 target values, but {target_classes_count} presented"
                self.logger.error(msg)
                raise ValidationError(msg)
        elif self.task_type == ModelTaskType.MULTICLASS:
            if not is_integer_dtype(target) and not is_string_dtype(target):
                if is_numeric_dtype(target):
                    try:
                        self[target_column] = self[target_column].astype("int")
                    except ValueError:
                        self.logger.exception("Failed to cast target to integer for multiclass task type")
                        raise ValidationError(
                            f"Unexpected dtype of target for multiclass task type: {target.dtype}."
                            "Expected int or str"
                        )
                else:
                    msg = f"Unexpected dtype of target for multiclass task type: {target.dtype}. Expected int or str"
                    self.logger.exception(msg)
                    raise ValidationError(msg)
        elif self.task_type == ModelTaskType.REGRESSION:
            if not is_float_dtype(target):
                try:
                    self[target_column] = self[target_column].astype("float")
                except ValueError:
                    self.logger.exception("Failed to cast target to float for regression task type")
                    raise ValidationError(
                        f"Unexpected dtype of target for regression task type: {target.dtype}. Expected float"
                    )
        elif self.task_type == ModelTaskType.TIMESERIES:
            if not is_float_dtype(target):
                try:
                    self[target_column] = self[target_column].astype("float")
                except ValueError:
                    self.logger.exception("Failed to cast target to float for timeseries task type")
                    raise ValidationError(
                        f"Unexpected dtype of target for timeseries task type: {target.dtype}. Expected float"
                    )

    def __resample(self):
        # self.logger.info("Resampling etalon")
        # Resample imbalanced target. Only train segment (without eval_set)
        if self.task_type in [ModelTaskType.BINARY, ModelTaskType.MULTICLASS]:
            if EVAL_SET_INDEX in self.columns:
                train_segment = self[self[EVAL_SET_INDEX] == 0]
                validation_segment = self[self[EVAL_SET_INDEX] != 0]
            else:
                train_segment = self
                validation_segment = None

            count = len(train_segment)
            min_class_count = count
            min_class_value = None
            target_column = self.etalon_def_checked.get(FileColumnMeaningType.TARGET.value, "")
            target = train_segment[target_column]
            target_classes_count = target.nunique()

            if target_classes_count > self.MAX_MULTICLASS_CLASS_COUNT:
                msg = (
                    f"The number of target classes {target_classes_count} exceeds the allowed threshold: "
                    f"{self.MAX_MULTICLASS_CLASS_COUNT}. Please, correct your data and try again"
                )
                self.logger.error(msg)
                raise ValidationError(msg)

            unique_target = target.unique()
            for v in list(unique_target):  # type: ignore
                current_class_count = len(train_segment.loc[target == v])
                if current_class_count < min_class_count:
                    min_class_count = current_class_count
                    min_class_value = v

            if min_class_count < self.MIN_TARGET_CLASS_COUNT:
                msg = (
                    f"The rarest class `{min_class_value}` occurs {min_class_count}. "
                    "The minimum number of observations for each class in a train dataset must be "
                    f"grater than {self.MIN_TARGET_CLASS_COUNT}. Please, correct your data and try again"
                )
                self.logger.error(msg)
                raise ValidationError(msg)

            min_class_percent = self.IMBALANCE_THESHOLD / target_classes_count
            min_class_threshold = min_class_percent * count

            if min_class_count < min_class_threshold:
                self.logger.info(
                    f"Target is imbalanced. The rarest class `{min_class_value}` occurs {min_class_count} times. "
                    "The minimum number of observations for each class in a train dataset must be "
                    f"grater than or equal to {min_class_threshold} ({min_class_percent * 100} %). "
                    "It will be undersampled"
                )

                if is_string_dtype(target):
                    target_replacement = {v: i for i, v in enumerate(unique_target)}  # type: ignore
                    prepared_target = target.replace(target_replacement)
                else:
                    prepared_target = target

                sampler = RandomUnderSampler(random_state=self.random_state)
                X = train_segment[SYSTEM_RECORD_ID]
                X = X.to_frame(SYSTEM_RECORD_ID)
                new_x, _ = sampler.fit_resample(X, prepared_target)  # type: ignore
                resampled_data = train_segment[train_segment[SYSTEM_RECORD_ID].isin(new_x[SYSTEM_RECORD_ID])]
                if validation_segment is not None:
                    resampled_data = pd.concat([resampled_data, validation_segment], ignore_index=True)
                self._update_inplace(resampled_data)
                self.logger.info(f"Shape after rebalance resampling: {self.shape}")
                self.sampled = True

        # Resample over fit threshold
        if EVAL_SET_INDEX in self.columns:
            train_segment = self[self[EVAL_SET_INDEX] == 0]
            validation_segment = self[self[EVAL_SET_INDEX] != 0]
        else:
            train_segment = self
            validation_segment = None
        if len(train_segment) > self.FIT_SAMPLE_THRESHOLD:
            self.logger.info(
                f"Etalon has size {len(train_segment)} more than threshold {self.FIT_SAMPLE_THRESHOLD} "
                f"and will be downsampled to {self.FIT_SAMPLE_ROWS}"
            )
            resampled_data = train_segment.sample(n=self.FIT_SAMPLE_ROWS, random_state=self.random_state)
            if validation_segment is not None:
                resampled_data = pd.concat([resampled_data, validation_segment], ignore_index=True)  # type: ignore
            self._update_inplace(resampled_data)
            self.logger.info(f"Shape after threshold resampling: {self.shape}")
            self.sampled = True

    def __convert_phone(self):
        """Convert phone/msisdn to int"""
        # self.logger.info("Convert phone to int")
        msisdn_column = self.etalon_def_checked.get(FileColumnMeaningType.MSISDN.value)
        if msisdn_column is not None and msisdn_column in self.columns:
            # self.logger.info(f"going to apply phone_to_int for column {msisdn_column}")
            phone_to_int(self, msisdn_column)
            self[msisdn_column] = self[msisdn_column].astype("Int64")

    def __features(self):
        return [
            f for f, meaning_type in self.meaning_types_checked.items() if meaning_type == FileColumnMeaningType.FEATURE
        ]

    def __remove_dates_from_features(self):
        # self.logger.info("Remove date columns from features")

        removed_features = []
        for f in self.__features():
            if is_datetime(self[f]) or is_period_dtype(self[f]):
                removed_features.append(f)
                self.drop(columns=f, inplace=True)
                del self.meaning_types_checked[f]

        if removed_features:
            msg = (
                f"Columns {removed_features} is a datetime or period type "
                "but not used as a search key and has been droped from X"
            )
            print(msg)
            self.logger.warning(msg)

    def __remove_empty_and_constant_features(self):
        # self.logger.info("Remove almost constant and almost empty columns")
        removed_features = []
        for f in self.__features():
            value_counts = self[f].value_counts(dropna=False, normalize=True)
            # most_frequent_value = value_counts.index[0]
            most_frequent_percent = value_counts.iloc[0]
            if most_frequent_percent >= 0.99:
                removed_features.append(f)
                self.drop(columns=f, inplace=True)
                del self.meaning_types_checked[f]

        if removed_features:
            msg = f"Columns {removed_features} has value with frequency more than 99% " "and has been droped from X"
            print(msg)
            self.logger.warning(msg)

    def __remove_high_cardinality_features(self):
        # self.logger.info("Remove columns with high cardinality")

        count = len(self)
        removed_features = []
        for f in self.__features():
            if (is_string_dtype(self[f]) or is_integer_dtype(self[f])) and self[f].nunique() / count >= 0.9:
                removed_features.append(f)
                self.drop(columns=f, inplace=True)
                del self.meaning_types_checked[f]
        if removed_features:
            msg = (
                f"Columns {removed_features} has high cardinality (>90% unique values) "
                "and has been droped from X"
            )
            print(msg)
            self.logger.warning(msg)

    def __validate_features_count(self):
        if len(self.__features()) > self.MAX_FEATURES_COUNT:
            msg = f"Maximum count of features is {self.MAX_FEATURES_COUNT}"
            self.logger.error(msg)
            raise Exception(msg)

    def __convert_features_types(self):
        # self.logger.info("Convert features to supported data types")

        for f in self.__features():
            if self[f].dtype == object:
                self[f] = self[f].astype(str)
            elif not is_numeric_dtype(self[f].dtype):
                self[f] = self[f].astype(str)

    def __validate_dataset(self, validate_target: bool, silent_mode: bool):
        """Validate DataSet"""
        # self.logger.info("validating etalon")
        date_millis = self.etalon_def_checked.get(FileColumnMeaningType.DATE.value) or self.etalon_def_checked.get(
            FileColumnMeaningType.DATETIME.value
        )
        target = self.etalon_def_checked.get(FileColumnMeaningType.TARGET.value)
        score = self.etalon_def_checked.get(FileColumnMeaningType.SCORE.value)
        if validate_target:
            if target is None:
                raise ValidationError("Target column is absent in meaning_types")

            if self.task_type != ModelTaskType.MULTICLASS:
                target_value = self.__target_value()
                target_items = target_value.nunique()
                if target_items == 1:
                    raise ValidationError("Target contains only one distinct value")
                elif target_items == 0:
                    raise ValidationError("Target contains only NaN or incorrect values.")

                self[target] = self[target].apply(pd.to_numeric, errors="coerce")
        keys_to_validate = [key for search_group in self.search_keys_checked for key in search_group]
        mandatory_columns = [date_millis, target, score]
        columns_to_validate = mandatory_columns.copy()
        columns_to_validate.extend(keys_to_validate)
        columns_to_validate = set([i for i in columns_to_validate if i is not None])

        nrows = len(self)
        validation_stats = {}
        self["valid_keys"] = 0
        self["valid_mandatory"] = True
        for col in columns_to_validate:
            self[f"{col}_is_valid"] = ~self[col].isnull()
            if validate_target and target is not None and col == target:
                self.loc[self[target] == np.Inf, f"{col}_is_valid"] = False

            if col in mandatory_columns:
                self["valid_mandatory"] = self["valid_mandatory"] & self[f"{col}_is_valid"]

            invalid_values = list(self.loc[self[f"{col}_is_valid"] == 0, col].head().values)  # type: ignore
            valid_share = self[f"{col}_is_valid"].sum() / nrows
            validation_stats[col] = {}
            optional_drop_message = "Invalid rows will be dropped. " if col in mandatory_columns else ""
            if valid_share == 1:
                valid_status = "All valid"
                valid_message = "All values in this column are good to go"
            elif 0 < valid_share < 1:
                valid_status = "Some invalid"
                valid_message = (
                    f"{100 * (1 - valid_share):.5f}% of the values of this column failed validation "
                    f"{optional_drop_message}"
                    f"Some examples of invalid values: {invalid_values}"
                )
            else:
                valid_status = "All invalid"
                valid_message = (
                    f"{100 * (1 - valid_share):.5f}% of the values of this column failed validation "
                    f"{optional_drop_message}"
                    f"Some examples of invalid values: {invalid_values}"
                )
            validation_stats[col]["valid_status"] = valid_status
            validation_stats[col]["valid_message"] = valid_message

            if col in keys_to_validate:
                self["valid_keys"] = self["valid_keys"] + self[f"{col}_is_valid"]
            self.drop(columns=f"{col}_is_valid", inplace=True)

        self["is_valid"] = self["valid_keys"] > 0
        self["is_valid"] = self["is_valid"] & self["valid_mandatory"]
        self.drop(columns=["valid_keys", "valid_mandatory"], inplace=True)

        drop_idx = self[self["is_valid"] != 1].index  # type: ignore
        self.drop(drop_idx, inplace=True)
        self.drop(columns=["is_valid"], inplace=True)

        if not silent_mode:
            df_stats = pd.DataFrame.from_dict(validation_stats, orient="index")
            df_stats.reset_index(inplace=True)
            df_stats.columns = ["Column name", "Status", "Description"]
            try:
                import html

                from IPython.display import HTML, display  # type: ignore

                def map_color(text):
                    colormap = {"All valid": "#DAF7A6", "Some invalid": "#FFC300", "All invalid": "#FF5733"}
                    return (
                        f"<td style='background-color:{colormap[text]};color:black'>{text}</td>"
                        if text in colormap
                        else f"<td>{text}</td>"
                    )

                df_stats["Description"] = df_stats["Description"].apply(lambda x: html.escape(x))
                html_stats = (
                    "<table>"
                    + "<tr>"
                    + "".join(f"<th style='font-weight:bold'>{col}</th>" for col in df_stats.columns)
                    + "</tr>"
                    + "".join("<tr>" + "".join(map(map_color, row[1:])) + "</tr>" for row in df_stats.itertuples())
                    + "</table>"
                )
                display(HTML(html_stats))
            except ImportError:
                print(df_stats)

    def __validate_meaning_types(self, validate_target: bool):
        # self.logger.info("Validating meaning types")
        if self.meaning_types is None or len(self.meaning_types) == 0:
            raise ValueError("Please pass the `meaning_types` argument before validation")

        if SYSTEM_RECORD_ID not in self.columns:
            self[SYSTEM_RECORD_ID] = self.apply(lambda row: hash(tuple(row)), axis=1)
            self.meaning_types[SYSTEM_RECORD_ID] = FileColumnMeaningType.SYSTEM_RECORD_ID

        for column in self.meaning_types:
            if column not in self.columns:
                raise ValueError(f"Meaning column {column} doesn't exist in dataframe columns: {self.columns}")
        if validate_target and FileColumnMeaningType.TARGET not in self.meaning_types.values():
            raise ValueError("Target column is not presented in meaning types. Specify it, please")

    def __validate_search_keys(self):
        # self.logger.info("Validating search keys")
        if self.search_keys is None or len(self.search_keys) == 0:
            raise ValueError("Please pass `search_keys` argument before validation")
        for keys_group in self.search_keys:
            for key in keys_group:
                if key not in self.columns:
                    raise ValueError(f"Search key {key} doesn't exist in dataframe columns: {self.columns}")

    def validate(self, validate_target: bool = True, silent_mode: bool = False):
        # self.logger.info("Validating dataset")

        self.__rename_columns()

        self.__validate_meaning_types(validate_target=validate_target)

        self.__validate_search_keys()

        self.__drop_ignore_columns()

        self.__remove_dates_from_features()

        self.__remove_empty_and_constant_features()

        self.__remove_high_cardinality_features()

        self.__validate_features_count()

        self.__validate_too_long_string_values()

        self.__clean_duplicates()

        self.__convert_bools()

        self.__convert_float16()

        self.__correct_decimal_comma()

        self.__to_millis()

        self.__remove_old_dates()

        self.__hash_email()

        self.__convert_ip()

        self.__convert_phone()

        self.__normalize_iso_code()

        self.__normalize_postal_code()

        self.__convert_features_types()

        self.__validate_dataset(validate_target, silent_mode)

        if validate_target:
            self.__validate_target()

            self.__resample()

            self.__validate_min_rows_count()

        self.__validate_max_row_count()

    def __construct_metadata(self) -> FileMetadata:
        # self.logger.info("Constructing dataset metadata")
        columns = []
        for index, (column_name, column_type) in enumerate(zip(self.columns, self.dtypes)):
            if column_name not in self.ignore_columns:
                if column_name in self.meaning_types_checked:
                    meaning_type = self.meaning_types_checked[column_name]
                    # Temporary workaround while backend doesn't support datetime
                    if meaning_type == FileColumnMeaningType.DATETIME:
                        meaning_type = FileColumnMeaningType.DATE
                else:
                    meaning_type = FileColumnMeaningType.FEATURE
                if meaning_type in {
                    FileColumnMeaningType.DATE,
                    FileColumnMeaningType.DATETIME,
                    FileColumnMeaningType.IP_ADDRESS,
                }:
                    min_max_values = NumericInterval(
                        minValue=self[column_name].astype("Int64").min(),
                        maxValue=self[column_name].astype("Int64").max(),
                    )
                else:
                    min_max_values = None
                column_meta = FileColumnMetadata(
                    index=index,
                    name=column_name,
                    originalName=self.columns_renaming.get(column_name) or column_name,
                    dataType=self.__get_data_type(column_type, column_name),
                    meaningType=meaning_type,
                    minMaxValues=min_max_values,
                )

                columns.append(column_meta)

        return FileMetadata(
            name=self.dataset_name,
            description=self.description,
            columns=columns,
            searchKeys=self.search_keys,
            hierarchicalGroupKeys=self.hierarchical_group_keys,
            hierarchicalSubgroupKeys=self.hierarchical_subgroup_keys,
            taskType=self.task_type,
        )

    def __get_data_type(self, pandas_data_type, column_name) -> DataType:
        if is_integer_dtype(pandas_data_type):
            return DataType.INT
        elif is_float_dtype(pandas_data_type):
            return DataType.DECIMAL
        elif is_string_dtype(pandas_data_type):
            return DataType.STRING
        else:
            msg = f"Unsupported data type of column {column_name}: {pandas_data_type}"
            self.logger.error(msg)
            raise Exception(msg)

    def __construct_search_customization(
        self,
        return_scores: bool,
        extract_features: bool,
        accurate_model: Optional[bool] = None,
        importance_threshold: Optional[float] = None,
        max_features: Optional[int] = None,
        filter_features: Optional[dict] = None,
        runtime_parameters: Optional[RuntimeParameters] = None,
    ) -> SearchCustomization:
        # self.logger.info("Constructing search customization")
        search_customization = SearchCustomization(
            extractFeatures=extract_features,
            accurateModel=accurate_model,
            importanceThreshold=importance_threshold,
            maxFeatures=max_features,
            returnScores=return_scores,
            runtimeParameters=runtime_parameters,
        )
        if filter_features:
            if [
                key
                for key in filter_features
                if key not in {"min_importance", "max_psi", "max_count", "selected_features"}
            ]:
                raise ValueError(
                    "Unknown field in filter_features. "
                    "Should be {'min_importance', 'max_psi', 'max_count', 'selected_features'}."
                )
            feature_filter = FeaturesFilter(
                minImportance=filter_features.get("min_importance"),
                maxPSI=filter_features.get("max_psi"),
                maxCount=filter_features.get("max_count"),
                selectedFeatures=filter_features.get("selected_features"),
            )
            search_customization.featuresFilter = feature_filter

        return search_customization

    def search(
        self,
        trace_id: str,
        return_scores: bool = False,
        extract_features: bool = False,
        accurate_model: bool = False,
        importance_threshold: Optional[float] = None,  # deprecated
        max_features: Optional[int] = None,  # deprecated
        filter_features: Optional[dict] = None,  # deprecated
        runtime_parameters: Optional[RuntimeParameters] = None,
    ) -> SearchTask:
        if self.etalon_def is None:
            self.validate()
        file_metrics = FileMetrics()

        file_metadata = self.__construct_metadata()
        search_customization = self.__construct_search_customization(
            return_scores=return_scores,
            extract_features=extract_features,
            accurate_model=accurate_model,
            importance_threshold=importance_threshold,
            max_features=max_features,
            filter_features=filter_features,
            runtime_parameters=runtime_parameters,
        )

        if self.file_upload_id is not None and get_rest_client(self.endpoint, self.api_key).check_uploaded_file_v2(
            trace_id, self.file_upload_id, file_metadata
        ):
            search_task_response = get_rest_client(self.endpoint, self.api_key).initial_search_without_upload_v2(
                trace_id, self.file_upload_id, file_metadata, file_metrics, search_customization
            )
        else:
            with tempfile.TemporaryDirectory() as tmp_dir:
                parquet_file_path = f"{tmp_dir}/{self.dataset_name}.parquet"
                self.to_parquet(path=parquet_file_path, index=False, compression="gzip")
                self.logger.info(f"Size of prepared uploading file: {Path(parquet_file_path).stat().st_size}")
                time.sleep(1)  # this is neccesary to avoid requests rate limit restrictions
                search_task_response = get_rest_client(self.endpoint, self.api_key).initial_search_v2(
                    trace_id, parquet_file_path, file_metadata, file_metrics, search_customization
                )
                self.file_upload_id = search_task_response.file_upload_id

        search_task = SearchTask(
            search_task_response.search_task_id,
            self,
            return_scores,
            extract_features,
            accurate_model,
            task_type=self.task_type,
            endpoint=self.endpoint,
            api_key=self.api_key,
        )
        return search_task.poll_result(trace_id)

    def validation(
        self,
        trace_id: str,
        initial_search_task_id: str,
        return_scores: bool = True,
        extract_features: bool = False,
        runtime_parameters: Optional[RuntimeParameters] = None,
        silent_mode: bool = False,
    ) -> SearchTask:
        if self.etalon_def is None:
            self.validate(validate_target=False, silent_mode=silent_mode)
        file_metrics = FileMetrics()

        file_metadata = self.__construct_metadata()
        search_customization = self.__construct_search_customization(
            return_scores, extract_features, runtime_parameters=runtime_parameters
        )

        if self.file_upload_id is not None and get_rest_client(self.endpoint, self.api_key).check_uploaded_file_v2(
            trace_id, self.file_upload_id, file_metadata
        ):
            search_task_response = get_rest_client(self.endpoint, self.api_key).validation_search_without_upload_v2(
                trace_id, self.file_upload_id, initial_search_task_id, file_metadata, file_metrics, search_customization
            )
        else:
            with tempfile.TemporaryDirectory() as tmp_dir:
                parquet_file_path = f"{tmp_dir}/{self.dataset_name}.parquet"
                self.to_parquet(path=parquet_file_path, index=False, compression="gzip")
                self.logger.info(f"Size of uploading file: {Path(parquet_file_path).stat().st_size}")
                time.sleep(1)
                search_task_response = get_rest_client(self.endpoint, self.api_key).validation_search_v2(
                    trace_id,
                    parquet_file_path,
                    initial_search_task_id,
                    file_metadata,
                    file_metrics,
                    search_customization,
                )
                self.file_upload_id = search_task_response.file_upload_id

        search_task = SearchTask(
            search_task_response.search_task_id,
            self,
            return_scores,
            extract_features,
            initial_search_task_id=initial_search_task_id,
            endpoint=self.endpoint,
            api_key=self.api_key,
        )

        return search_task.poll_result(trace_id, quiet=silent_mode)



import json

import requests

try:
    from packaging.version import parse
except ImportError:
    from pip._vendor.packaging.version import parse

import logging

from upgini.http import __version__

URL_PATTERN = "https://pypi.python.org/pypi/{package}/json"


def get_version(package, url_pattern=URL_PATTERN):
    """Return version of package on pypi.python.org using json."""
    req = requests.get(url_pattern.format(package=package))
    version = parse("0")
    if req.status_code == requests.codes.ok:
        j = json.loads(req.text.encode(req.encoding or "utf8"))
        releases = j.get("releases", [])
        for release in releases:
            ver = parse(release)
            if not ver.is_prerelease:
                version = max(version, ver)
    return version


def validate_version(logger: logging.Logger):
    try:
        current_version = parse(__version__)
        latest_version = get_version("upgini")
        if current_version < latest_version:  # type: ignore
            msg = f"You use {current_version} version, but latest is {latest_version}"
            logger.warning(msg)
            print("WARNING: " + msg)
    except Exception:
        logger.exception("Failed to validate version")
